#!/usr/bin/env python3
import itertools as itt
import numpy as np
import sys

class Paramgroup(object):
    def __init__(self, pg={}, rep={}):
        self._groups = []
        self._repeats = {}
        self._product_groups = {}  # pg[0]['max'] = [50, 100, 150]
        for gr, items in pg.items():
            for name, params in items.items():
                self.add(gr, name, params)
        for gr, reps in rep.items():
            self.set_repeats(gr, reps)

        self._defaults = {}

        self._seedname = "general:seed"
        self._run_name = "run"
        self._groupname = "group"
        self._setname = "set"

        self._run_name = "general:run"
        self._groupname = "general:group"
        self._setname = "general:set"
        self.default('general:output_frequency', 1000000)

    def _finalize(self):
        for group in self.groups:
            for name, params in self._defaults.items():
                if name not in self._product_groups[group]:
                    self.add(group, name, params)

    def __iter__(self):
        self._finalize()
        header = self.header[3:]  # TODO: header
        count = 0
        for group in self.groups:
            for _ in range(self._repeats[group]):
                vals = [self._product_groups[group][key] for key in header]
                for ps_i, param_set in enumerate(itt.product(*vals), 1):
                    count += 1
                    yield (count, group, ps_i) + param_set

    @property
    def groups(self):
        return self._groups.copy()

    @property  # TODO: make this cleaner and not suck
    def header(self):
        header = self._get_header_from_group(self.groups[0])
        for group in self.groups[1:]:
            this = self._get_header_from_group(group)
            if this != header:
                msg = "Headers for group ({}) do not match:\n{}\n{}"
                msg = msg.format(group, header, this)
                raise RuntimeError(msg)
        cols_prefix = [self._run_name, self._groupname, self._setname]
        return cols_prefix + header

    def print(self, outfile=None, logfile=None, delimiter=","):
        if len(self.groups) == 0:
            raise RuntimeError("No parameters added.")

        self._finalize()
        header = self.header
        with open(outfile, "w") if outfile else sys.stdout as of:
            print("#", self, file=of)
            print(*header, self._seedname, sep=delimiter, file=of)

            porder = [header.index(name) for name in header]
            for param_set in self:
                line = [param_set[i] for i in porder]
                print(*line, np.random.randint(1e8), sep=delimiter, file=of)

        if logfile and outfile:
            with open(logfile, "a") as lf:
                from datetime import date

                print(date.today().strftime("%Y.%m.%d"), outfile, self, file=lf)

    def _get_header_from_group(self, group):
        header = list(sorted(self._product_groups[group].keys()))
        if len(header) != len(set(header)):
            msg = "One or more param names repeated: {}".format(header)
            raise RuntimeError(msg)
        return header

    @staticmethod
    def _tolist(params):
        try:
            params = params.tolist()  # numpy
        except AttributeError:
            if not isinstance(params, str) and not hasattr(params, "__iter__"):
                params = [params]
        return params

    def default(self, name, param):
        param = self._tolist(param)
        self._defaults[name] = param

    def add(self, group, name, params, override=False):
        params = self._tolist(params)

        if group not in self.groups:
            self._groups.append(group)
            self._repeats[group] = 1
            self._product_groups[group] = {}

        if not override and name in self._product_groups[group]:
            print(f"Warning: Overwriting {name} in {group}")
        self._product_groups[group][name] = params

    def set_repeats(self, n, group=None):
        if group:
            if group not in self.groups:
                raise RuntimeError("{} not in groups: {}".format(group, self.groups))
            self._repeats[group] = n
        else:
            for gr in self.groups:
                self._repeats[gr] = n

    def __repr__(self):
        return "Paramgroup(pg={}, rep={})".format(self._product_groups, self._repeats)

    def __len__(self):
        self._finalize()
        gc = {
            gr: np.prod([len(x) for x in self._product_groups[gr].values()])
            for gr in self.groups
        }
        return sum(self._repeats[gr] * gc[gr] for gr in self.groups)


def main():
    import argparse

    parser = argparse.ArgumentParser()
    # TODO: make this appropriately optional
    # parser.add_argument("set", metavar="Parameter set", help="Name of function to call")
    parser.add_argument(
        "set", metavar="Parameter set", nargs="?", help="Name of function to call"
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-o", metavar="output_file")
    group.add_argument("-e", metavar=("header_file"))
    group.add_argument("-n", metavar=("logfile line_number"))
    group.add_argument("-p", metavar=("Param line"), nargs="+")
    parser.add_argument(
        "-v", metavar=("video_reps", "output_frequency"), nargs=2, type=int
    )
    parser.add_argument("-l", metavar="logfile", default="params.log")
    parser.add_argument("-d", metavar="delimiter", default=",")
    parser.add_argument("-c", action="count", help="Print count")

    args = parser.parse_args()
    if not args.set and not args.e and not args.n and not args.p:
        parser.print_help()
        print(
            "\nParameter set, header file, log file, or parameter line number required."
        )
        return

    if args.p:
        pg = Paramgroup(**_parse_line(" ".join(args.p)))
        pg.print(delimiter=args.d)
    elif args.e:
        with open(args.e, "r") as header_file:
            header = next(header_file)
        pg = Paramgroup(**_parse_line(header))
        pg.print(delimiter=args.d)
    elif args.n:
        with open(args.l, "r") as logfile:
            for i, line in enumerate(logfile, 1):
                if i == int(args.n):
                    break
            else:
                raise RuntimeError("{} does not have {} lines.".format(args.l, args.n))
            pg = Paramgroup(**_parse_line(line))
            pg.print(delimiter=args.d)
    else:
        pg = parameter_sets[args.set]()
        if args.v:
            pg = videos(pg, *args.v)
        if args.c:
            print(len(pg))
        else:
            pg.print(args.o, args.l, delimiter=args.d)


def _parse_line(line):
    import re
    from ast import literal_eval

    pg = literal_eval(re.sub(".*pg=(\{.*\}), rep=.*", r"\1", line))
    rep = literal_eval(re.sub(".*pg=.*, rep=(\{.*\})\)$", r"\1", line))
    return {"pg": pg, "rep": rep}


def test():
    pg = Paramgroup()

    pg.add("group1", "param1", 7)
    pg.add("group1", "product1", [1, 2, 3])
    pg.set_repeats(3, "group1")

    pg.add("group2", "param1", 5)
    pg.add("group2", "product1", [2, 3])

    pg.add("group3", "param1", 6)
    pg.add("group3", "product1", [8, 9])
    pg.set_repeats(2, "group3")
    return pg


def videos(pg, reps=1, freq=1):
    # for gr in pg:  # TODO
    for gr in pg.groups:
        pg.set_repeats(reps, gr)
        pg.add(gr, "general:output_frequency", freq, override=True)

    return pg


def pgset_well_mixed():
    # TODO: defaults. same as looping through groups..
    # TODO: print count of params
    pg = Paramgroup()

    gr = "low"
    pg.add(gr, "general:infection_start", 7)
    pg.add(gr, "substrate:max", 0.043)
    pg.add(gr, "general:runtime", 17)

    gr = "mid"
    pg.add(gr, "general:infection_start", 2)
    pg.add(gr, "substrate:max", 0.185)
    pg.add(gr, "general:runtime", 12)

    gr = "max"
    pg.add(gr, "general:infection_start", 2)
    pg.add(gr, "substrate:max", 0.2)
    pg.add(gr, "general:runtime", 12)

    for gr in pg.groups:
        pg.add(gr, "general:init_ratio", [0.97, 0.65, 0.35, 0.03])
        pg.add(gr, "general:slow", [0.0])
        pg.add(gr, "general:resistance_cost", [0, 0.05, 0.35, 0.65])
        pg.add(gr, "space:well_mixed", True)

    pg.set_repeats(6)
    return pg


parameter_sets = {}


def parameterset(func):
    global parameter_sets
    parameter_sets[func.__name__] = func
    return parameter_sets[func.__name__]


@parameterset
def examplename():
    # TODO: print count of params
    pg = Paramgroup()
    # pg.default("general:runtime", 20)
    # pg.default("general:targetone", True)
    # pg.default("general:init_count", 150)
    # pg.default("general:init_f", 1.0)
    # pg.default("general:fixed_dt", 0)
    # pg.default("infection:count", 1)    # TO DO change phage blanket function for init phage sweep
    # pg.default("infection:duration", 5)
    # pg.default("phage:burst", 120)
    # pg.default("substrate:max", 6)
    pg.default("general:output_frequency", 1)

    gr = 'group1'
    pg.add(gr, "species1:impedance", [3.0, 4.5, 6.0, 7.5, 9.0])
    pg.add(gr, "infection:init_phage", [10, 110, 220, 330, 450])
    #pg.add(gr, "substrate:max", [1.0, 2.0, 4.0, 6.0])
    # gr = 'group2'
    # pg.add(gr, "erosion:phage_rate", 2.4e12)  # has no default, must be added if it's in another group
    # pg.add(gr, "infection:height", [10e-6, 60e-6, 90e-6])

    pg.set_repeats(3)       # TODO change to 6

    return pg


if __name__ == "__main__":
    # usage example: ./make_parameters.py examplename -c
    # usage example: ./make_parameters.py examplename -o filename.csv
    # usage example: ./make_parameters.py examplename -o filename.csv -v 1 1
    main()
