import pandas
from time import time
import numpy as np
import multiprocessing as mp
from re import sub, match
import warnings
import argparse
import os
import csv
from contextlib import contextmanager

import matplotlib
from matplotlib import cm
import matplotlib.colors
from matplotlib.colors import cnames, hex2color, rgb_to_hsv, hsv_to_rgb
import matplotlib.animation as animation
from matplotlib.patches import Patch
import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d
from matplotlib.widgets import Slider, Button, RadioButtons

#from tqdm import tqdm
from itertools import product
from functools import partial

# matplotlib.use("agg")
#tqdm80 = partial(tqdm, ncols=80)
# matplotlib deprecation warnings
warnings.filterwarnings("ignore", category=UserWarning)

import plotnine
from plotnine import scales, geoms, labels, guides, guide_legend
from plotnine import ggplot, aes, facet_wrap, stats
from plotnine import element_text, theme, coords
from plotnine.themes import theme_bw, theme_classic

# import vapory_git as pov
# from vapory_git import Camera, Scene, Background, LightSource, Texture, Pigment
# from vapory_git import Box, Sphere, Finish, Normal

import simbiofilm as sb

from tempfile import mkstemp

#import cv2
import glob


@contextmanager
def ignore_copywarn():
    pandas.options.mode.chained_assignment = None
    yield
    pandas.options.mode.chained_assignment = "warn"


def _plot_frame(
    file,
    outdir,
    shape,
    crop=None,
    max_biomass=6.5e-12,
    max_phage=1000,
    format="png",
    prefix="",
    style=None,
    solute_adjustment=None,
):

    warnings.filterwarnings(
        "ignore", category=UserWarning
    )  # matplotlib deprecation warnings

    outname = f'{outdir}/{prefix}{file.split("/")[-1][:-4]}.{format}'
    with np.load(file, allow_pickle=True) as f:
        biomass_names = [x for x in f.keys() if "_data" in x and "phage" not in x]
        phage_names = [x for x in f.keys() if "_data" in x and "phage" in x]

        gridcols = np.unravel_index(range(np.prod(shape)), shape)

        # set up biomass
        biomass = [np.rec.array(f[x]) for x in biomass_names]
        biomass_totals = [np.zeros(np.prod(shape)) for x in biomass_names]
        [np.add.at(tot, c.location, c.mass) for tot, c in zip(biomass_totals, biomass)]

        def _make_df(data):
            return pandas.DataFrame({"count": data, "x": gridcols[1], "y": gridcols[0]})

        dfs = [_make_df(d) for d in biomass_totals]
        for df, species in zip(dfs, biomass_names):
            df["species"] = sub("_data", "", species)
        bio = pandas.concat(dfs)
        minv = 0.20
        minb = bio[bio["count"] > 0]["count"].min()
        bio["alpha"] = ((0.95 - minv) * (bio["count"] - minb)) / (
            max_biomass - minb
        ) + minv

        # phage
        # phage = np.rec.array(f[phage_names[0]])
        phage_total = np.zeros(np.prod(shape))
        # np.add.at(phage_total, phage.location, 1)
        phg = _make_df(phage_total)

        minv = 0.250
        if max_phage > 0:
            max_phage = np.log10(max_phage)
        minp = np.log10(phg[phg["count"] > 0]["count"].min())
        if max_phage <= minp:
            max_phage += min(0.01, max_phage * 1.1)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")  # ignore log10(0) warnings
            phg["alpha"] = ((0.95 - minv) * (np.log10(phg["count"]) - minp)) / (
                max_phage - minp
            ) + minv

        phg["species"] = "phage"

        # combine, and clear out empty cells. Otherwise, alpha is minv

        # NORMAL:
        colors = ["red", "blue", "green", "black"]
        lims = ["species1", "species2", "infected", "phage"]

        # if style == "single":
        #     colors = ["red", "orange", "green", "black"]
        #     lims = ["Species1", "Matrix1", "infected", "phage"]

        # if style == "two":
        #     colors = ["red", "blue", "orange", "#aa7711", "green", "black"]
        #     lims = ["Species1", "Species2", "Matrix1", "Matrix2", "infected", "phage"]

        # colors = ["red", "blue", "orange"]
        # lims = ["Species1", "Species2", "Matrix1"]

        if solute_adjustment is not None:
            sv = np.log10(f["solute_value"] * 1 + solute_adjustment)
            mins = sv[sv.nonzero()].min()
            maxs = sv[sv.nonzero()].max()
            sv = ((0.95) * (sv - mins)) / (maxs - mins)
            solute = pandas.DataFrame(
                {
                    "y": gridcols[0],
                    "x": gridcols[1],
                    "alpha": sv[gridcols[0], gridcols[1]],
                    "species": "substrate",
                    "count": f["solute_value"][gridcols[0], gridcols[1]],
                }
            )

            data = pandas.concat([bio, phg, solute], sort=False)
            # colors.append('#ffffff')
            colors.append("#3c1978")
            lims.append("substrate")
        else:
            data = pandas.concat([bio, phg], sort=False)

        data = data[data["count"] > 0]
        if crop is not None:
            data = data[(data.x >= crop[1][0]) & (data.x < crop[1][1])].copy()
            data["x"] -= crop[1][0]
            shape = (crop[0], crop[1][1] - crop[1][0])

        # # lims = ['susceptible', 'resistant', 'infected', 'phage']
        # sections = np.unique([x.split(':')[0] for x in f['config'].T[0]])
        # if any(['matrix' in x for x in sections]):
        #     colors.append('yellow')
        #     lims.append('matrix')
        #     for name in [x for x in data.species.unique() if 'matrix' in x]:
        #         data.loc[data.species == name, 'species'] = 'matrix'

        dl = 3
        # dl = int(cfg.space.dl * 1e6)
        # bio.loc[bio.species == 'Species1', 'species'] = 'susceptible'
        # bio.loc[bio.species == 'Species2', 'species'] = 'resistant'


        timestr = "{:.2f} days".format(f["t"]).rjust(10, " ")
        data["y"] += 1
        plot = ggplot()
        plot += geoms.geom_tile(
            aes(x="x", y="y", alpha="alpha", fill="factor(species)"), data=data
        )
        plot += scales.scale_alpha_continuous(range=[0, 1], limits=[0, 1], guide=False)
        plot += scales.scale_fill_manual(values=colors, limits=lims, drop=False)
        plot += labels.ggtitle("t =" + timestr)
        plot += scales.scale_y_continuous(
            limits=[0, shape[0]],
            expand=[0, 0],
        )
        plot += scales.scale_x_continuous(
            limits=[0, shape[1]],
            expand=[0, 0],
        )
        plot += guides(fill=guide_legend(title=""))
        plot += labels.xlab("") + labels.ylab("")
        plot += theme_classic()
        plot += theme(plot_title=element_text(size=8))
        wid = 5.0
        hi = wid * shape[0] / shape[1]
        plot.save(
            outname, dpi=300, verbose=False, limitsize=False
        )


def noframe(f, outdir):
    outname = "{}/{}.png".format(outdir, f[:-4])
    return not os.path.isfile(outname)


def plot_all_frames(
    datadir,
    outdir,
    shape,
    crop=None,
    maxp=None,
    maxb=None,
    format="png",
    style=None,
    nprocesses=4,
    phage_time=False,
    solute_adjustment=None,
    **kwargs,
):
    print(outdir)
    if not os.path.exists(outdir):
        os.makedirs(outdir)
        
    for x in os.listdir(datadir):
        print(x)
    flist = [
        "{}/{}".format(datadir, x)
        for x in os.listdir(datadir)
        if ".npz" in x and noframe(x, outdir)
    ]
    if not flist:
        print('hoi')
        return
    # cfg = sb.getcfg(flist[0])
    # shape = tuple(map(int, cfg.space.shape))

    if len(shape) == 2:
        maxb, maxp = 6.5e-12, 1000
        # solute_adjustment = 0.05
        if nprocesses > 1:
            if (maxb is None and maxp is None) or phage_time:
                with mp.Pool(nprocesses) as p:
                    maxes = p.starmap(_get_max, [(fname, shape) for fname in flist])
                    if maxb is None:
                        maxb = max(x[0] for x in maxes)
                    if maxp is None:
                        maxp = int(max(x[1] for x in maxes))
                    if phage_time:
                        return

            with warnings.catch_warnings(), mp.Pool(nprocesses) as p:
                warnings.simplefilter("ignore")  # ignore log10(0) warnings
                p.starmap(
                    _plot_frame,
                    [
                        (
                            f,
                            outdir,
                            shape,
                            crop,
                            maxb,
                            maxp,
                            format,
                            "",
                            style,
                            solute_adjustment,
                        )
                        for f in flist
                    ],
                )
        else:
            if maxb is None and maxp is None:
                maxes = [_get_max(fname, shape) for fname in flist]
                if maxb is None:
                    maxb = max(x[0] for x in maxes)
                if maxp is None:
                    maxp = int(max(x[1] for x in maxes))
            for f in flist:
                print(f)
                _plot_frame(
                    f,
                    outdir,
                    shape,
                    crop,
                    maxb,
                    maxp,
                    format,
                    style=style,
                    solute_adjustment=solute_adjustment,
                )
    if len(shape) == 3:
        print('yes')
        maxb, maxp = 6.5e-12, 1000
        if nprocesses > 1:
            with warnings.catch_warnings(), mp.Pool(nprocesses) as p:
                warnings.simplefilter("ignore")  # ignore log10(0) warnings
                p.starmap(
                    _render_3d_frame,
                    [(f, outdir, shape, maxb, maxp, format, "") for f in flist],
                )
        else:
            for f in flist:
                _render_3d_frame(f, outdir, shape, 6.5e-12, 1000, format, "")


def _get_max(fname, shape):
    maxb, maxp = 0, 0
    total = np.zeros(np.prod(shape))
    with np.load(fname) as f:
        time = f["t"]
        for name in [x for x in f.keys() if "_data" in x]:
            total[:] = 0
            if "phage" in name:
                np.add.at(total, f[name]["location"], 1)
                maxp = max([maxp, total.max()])
            else:
                np.add.at(total, f[name]["location"], f[name]["mass"])
                maxb = max([maxb, total.max()])
    return time, maxb, maxp


def load_data(
    datadir,
    header=1,
    write_summaries=True,
    force_load=False,
    funcs=None,
    dtypes={},
    return_cfgs=False,
):
    """ Funcs is for additional processing before the file gets written, so it doesn't
    need to be repeated in the future. """
    print("Loading cfg...", flush=True)
    cfgs = pandas.read_csv(datadir + "/parameters.csv", header=header, dtype="category")
    # cfgs.insert(0, "general:run", range(1, len(cfgs) + 1))
    cols = [rc.split(":")[1] for rc in cfgs.columns]
    if len(cols) > len(set(cols)):
        fcols = [rc.replace(":", "_") for rc in cfgs.columns]
        cols = [f if cols.count(c) > 1 else c for f, c in zip(fcols, cols)]
    cfgs.columns = cols
    cfgs.run = cfgs.run.astype(int)
    cfgs = cfgs.set_index("run").drop(columns='output_frequency')

    print("Loading summaries...", flush=True)
    feathersummary = datadir + "/full_summary.feather"
    if not os.path.isfile(feathersummary) or force_load:
        summaries = _load_summaries(datadir, np.array(cfgs.index))

        print("Adding cfg to summaries...", flush=True)
        summaries.index = summaries.index.astype(int).astype('category')
        df = cfgs[cfgs.index.isin(summaries.index)].combine_first(summaries)
        for k, v in dtypes.items():
            df[k] = df[k].astype(v).astype("category") if k in cfgs else df[k].astype(v)

        df.iteration = df.iteration.astype(int)
        if "site" in df:
            df["site"] = df.site.astype(int).astype("category")

        try:
            for f in funcs:
                df = f(df)
        except TypeError:  # Covers not callable or iterable
            if funcs:
                df = funcs(df)

        if write_summaries:
            try:
                print("Writing summaries to {}".format(feathersummary), flush=True)
                df.reset_index().to_feather(feathersummary)
            except:
                print("Tried and failed to write summaries.")

    else:
        df = pandas.read_feather(feathersummary)
        df = df.set_index("run")

    df.index = df.index.astype("category")
    return df if not return_cfgs else (df, cfgs)


def _load_summaries(datadir, runs):

    # def _loadone(filename, run, data, failed, simn=None):
    def _loadone(filename, simn=None):
        try:
            with open(filename) as summaryfile:
                run_summary = pandas.read_csv(summaryfile)
        except FileNotFoundError:
            failed.append(run if simn is None else f"{run}.{simn}")
            return
        run_summary["run"] = int(run)
        if simn is not None:
            run_summary["site"] = simn
        data.append(run_summary.set_index("run"))

    ##
    data = []
    failed = []
    multisim = len(
        [f for f in os.listdir(f"{datadir}/data/run_{1:06}") if match("^run\d+$", f)]
    )

    if multisim > 0:
        fnstring = f"{datadir}/data/run_{{:06}}/run{{}}/summary.csv"
    else:
        fnstring = f"{datadir}/data/run_{{:06}}/summary.csv"

    for run in tqdm80(runs):
        if multisim > 0:
            for i in range(1, multisim + 1):
                _loadone(fnstring.format(int(run), i), i)
        else:
            _loadone(fnstring.format(datadir, run))

    if failed:
        print("Unable to load: {}\n{} Total".format(failed, len(failed)))
    return pandas.concat(data, sort=True)


def _main():
    matplotlib.use("agg")
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--video")
    parser.add_argument("-f", "--format")
    parser.add_argument("--crop", nargs=3)
    parser.add_argument("--maxb")
    parser.add_argument("--maxp")
    parser.add_argument("--solute_adjustment")
    parser.add_argument("--interval")
    parser.add_argument("--name")
    parser.add_argument("--prefix")
    parser.add_argument("--progress", action="store_true")
    parser.add_argument("--skip", action="store_true")
    parser.add_argument("--jobs")
    parser.add_argument("-o", "--output")

    args = parser.parse_args()

    if args.video:

        if not args.format:
            raise RuntimeError("Video requires format argument (-f)")
        else:
            kwargs = {"format": args.format}

        if args.crop:
            kwargs["crop"] = [int(x) for x in args.crop]
        if args.maxb:
            kwargs["maxb"] = float(args.maxb)
        if args.maxp:
            kwargs["maxp"] = int(args.maxp)
        if args.solute_adjustment:
            kwargs["solute_adjustment"] = float(args.solute_adjustment)
        if args.interval:
            kwargs["interval"] = int(args.interval)
        if args.name:
            kwargs["name"] = args.name
        if args.prefix:
            kwargs["prefix"] = args.prefix
        if args.jobs:
            kwargs["nprocesses"] = int(args.jobs)
        if args.output:
            kwargs["outdir"] = args.output
        kwargs["progress_bar"] = args.progress
        kwargs["skip_made"] = args.skip

        plot_all_frames(args.video, **kwargs)
  
        
def plot_toxin_profile(outdir, *summary_files):

    warnings.filterwarnings(
        "ignore", category=UserWarning
    )  # matplotlib deprecation warnings

    outname = f'{outdir}/toxin_profile'
    t_list = []
    
    tox_list = []
    
    for summary_file in summary_files:
        print(summary_file)
        with open(summary_file, newline='') as csvfile:
            csvreader = csv.reader(csvfile, delimiter=' ')
            count = 0
            for row in csvreader:
                if count == 0:
                    count +=1
                    continue
                print(row[0].split(',')[-1])
                t_list.append(float(row[0].split(',')[1]))
                tox_list.append(float(row[0].split(',')[-1]))
    t_list = np.mean(np.array(t_list), axis=0)
    tox_list = np.mean(np.array(tox_list), axis=0)
  
    plt.clf()
    plt.figure(1)
    plt.plot(t_list, tox_list, '-b')
    plt.xlabel('time (days)')
    plt.ylabel('toxin (number of toxic molecules)')
    plt.savefig(outname)
    return

        # gridcols = np.unravel_index(range(np.prod(shape)), shape)

        # # set up biomass
        # biomass = [np.rec.array(f[x]) for x in biomass_names]
        # biomass_totals = [np.zeros(np.prod(shape)) for x in biomass_names]
        # [np.add.at(tot, c.location, c.mass) for tot, c in zip(biomass_totals, biomass)]

def make_video(datadir, outdir):
    img_array = []
    for filename in glob.glob(datadir+'*.png'):
        img = cv2.imread(filename)
        height, width, layers = img.shape
        size = (width,height)
        img_array.append(img)
    outpath = outdir + '/video.avi'
    out = cv2.VideoWriter(outpath,cv2.VideoWriter_fourcc(*'DIVX'), 15, size)
     
    for i in range(len(img_array)):
        out.write(img_array[i])
    out.release()

if __name__ == "__main__":
     # datadir = 'C:/Users/rdvan/igem2020/Model3/Simmons2020_adjusted/simbiofilm/phage/Simulation/'
     # outdir = 'C:/Users/rdvan/igem2020/Model3/Simmons2020_adjusted/local_runs/simbiofilm/examples/phage/Plots'
     # datadir = 'C:/Users/rdvan/igem2020/Model3/Simmons2020_adjusted/local_runs/simbiofilm/examples/phage/'
     # summary_file = 'C:/Users/rdvan/igem2020/Model3/Simmons2020_adjusted/local_runs/simbiofilm/examples/phage/summary.csv'
     
    for i in range(1,13):
        datadir = 'C:/Users/rdvan/iGEM_sweeps/Burst/sweep/data/run_0000' + str(i) +'/'
        summary_file = 'C:/Users/rdvan/iGEM_sweeps/Burst/sweep/data/run_0000' + str(i) + '/summary.csv'
        outdir = 'C:/Users/rdvan/iGEM_sweeps/Burst/sweep/data/run_0000' + str(i) + '/plots'
        plot_all_frames(datadir, outdir, (90,450))
       

    for i in range(1,6):
        sum_file1 = 'C:/Users/rdvan/iGEM_sweeps/Burst/sweep/data/run_{:06d}/summary.csv'.format(i)
        sum_file2 = 'C:/Users/rdvan/iGEM_sweeps/Burst/sweep/data/run_{:06d}/summary.csv'.format(i+5)
        sum_file3 = 'C:/Users/rdvan/iGEM_sweeps/Burst/sweep/data/run_{:06d}/summary.csv'.format(i+10)
        outdir = 'C:/Users/rdvan/iGEM_sweeps/Burst/sweep/data/toxin_profiles/{}'.format(i)
        if not os.path.exists(outdir):
            os.makedirs(outdir)
        plot_toxin_profile(outdir, sum_file1, sum_file2, sum_file3)
         
     #make_video(outdir, outdir)
     
                
